/*
 Battery Utility for Dingoo A320
	Rodrigo Cardoso - 2011/02/28
 
 License: Use whatever way you want :P
*/
   
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>


// I'm testing the interface in an Apple machine...
#if !defined (__APPLE__)

#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include <SDL/SDL_ttf.h>
#include <SDL/SDL_mixer.h>

#else

// The way I need to include these things in Xcode are strange =P
#include <SDL/SDL.h>
#include <SDL_image/SDL_image.h>
#include <SDL_ttf/SDL_ttf.h>
#include <SDL_mixer/SDL_mixer.h>

#endif

#include "dingux.h"

// TODO: ver se precisa iniciar a biblioteca antes de carregar um som,
// se der, carregar o som antes, e so carregar a biblioteca qdo for tocar.
const int initflags		= SDL_INIT_VIDEO;
const int video_w		= 320;
const int video_h		= 240;
const int video_bpp		= 16;
const int video_flags	= SDL_SWSURFACE;

const int timer_interval= 1000;
const int brightness_timer_interval= 10000;

#define TOLERANCE_RANGE	8

#define RGB(r,g,b)		SDL_MapRGB( screen->format, r, g, b )
#define EQUALS(a,b)		!strcmp(a, b)
#define CENTER(img)		((video_w - img->w) / 2)

#define CLAMP(a, min, max) \
	((a > max) ? max : ((a < min) ? min : a) )

// Lazyness macros
#define	CHECKSETPROPERTY(file, var, field, format)	\
	if (EQUALS(var, #field)) {\
		fscanf(file, format, &field);\
		continue;\
	}

#define	CHECKSETPROPERTYPOINT(file, var, field)	\
	if (EQUALS(var, #field)) {\
		fscanf(file, "%d %d", &field.x, &field.y);\
		continue;\
	}


typedef enum {
	FALSE = 0,
	TRUE = 1
} bool;


typedef struct {
	int x, y;
} point;



// load an image and convert it to a optimized format
SDL_Surface *loadImage( const char *filename, int isColorKeyed, Uint32 colorKey )
{
	SDL_Surface		*loadedImage	= NULL,
					*optimizedImage	= NULL;
	
	loadedImage = IMG_Load( filename );
	
	if( loadedImage != NULL )
	{
		optimizedImage = SDL_DisplayFormat( loadedImage );	//Create an optimized surface
		SDL_FreeSurface( loadedImage );						// Trash the old surface
		
		if( optimizedImage && isColorKeyed )				// Set a colorkey if necessary
		{
			Uint8			r, g, b;
			SDL_PixelFormat *format = optimizedImage->format;
			
			SDL_GetRGB(colorKey, format, &r, &g, &b); // Get r,g and b from an integer
			SDL_SetColorKey( optimizedImage, SDL_SRCCOLORKEY, SDL_MapRGB( format, r, g, b ) );
		}
	}
	else {
		fprintf(stderr, "Error Loading the image:\n\t%s\n", filename);
	}
	
	return optimizedImage;
}

// the blit function...
void applyBitmap( int x, int y, SDL_Surface* src, SDL_Surface* dst, SDL_Rect* clip )
{
	SDL_Rect offset;	//Holds offsets
	
	offset.x = x;		//Get offsets
	offset.y = y;
	
	SDL_BlitSurface( src, clip, dst, &offset ); // Blit
}



int main(int argc, char *argv[])
{
	/************/
	/* vars ... */
	
	SDL_Surface *screen = NULL;
	SDL_Event	e;
	
	FILE		*config = NULL,
				*themefile = NULL,
				*battinfo = NULL;
	
	SDL_Surface *bg_surf = NULL,
				*fill_surf = NULL,
				*completed_surf = NULL;
	
	Mix_Chunk	*beep = NULL;
	
	TTF_Font	*font = NULL; 
	
	int			done = 0,
				old_clock = get_clock(),
				default_brightness = get_brightness(),
				current_brightness = default_brightness,
				faded_brightness = 5,
				default_volume = get_vol();
	
	bool		battery_percent_changed = TRUE,
				brightness_fading = FALSE,
				sound_enabled = TRUE;
	
	// messages
	SDL_Surface *msg1 = NULL,
				*msg1_shadow = NULL,
				*msg2 = NULL,
				*msg2_shadow = NULL,
				*time = NULL,
				*time_shadow = NULL,
				*status = NULL,
				*status_shadow = NULL,
				*percent = NULL,
				*percent_shadow = NULL;
	
	// battery info
	int			battery_level = get_battery_level(), // more or less the minimum value
	
				battery_min = 3000, // common values
				battery_max = 4100,
				battery_calibrate = 0,
	
				battery_percent_old = -1,
				battery_percent = -2;
	
	// default options
	char		theme[64] = "";
	int			show_time = 0,
				show_pct = 0,
				show_msgs = 0,
				show_status = 0,
				beep_when_done = 0,
				proc_clock = 0,
				auto_brightness = 0,
				bright_pulse = 0;
	
	// theme parameters
	point		bar_pos = { 0, 0 },
				completed_pos = { 0, 0 },
				percent_pos = { 0, 20 },
				msg1_pos = { 0, 40 },
				msg2_pos = { 0, 60 },
				status_pos = { 0, 80 },
				time_pos = { 0, 100 };
	
	int			text_size = 10;
	SDL_Color	text_color = { 255, 255, 255 },
				text_shadow = { 0, 0, 0 };

	/********************/
	/* file handling... */
	
	// config file
	config = fopen("config.txt", "r");
	if (config == NULL) {
		fprintf(stderr, "File config.txt not found! Creating one.\n");
		config = fopen("config.txt", "w+");
		
		if (config == NULL) {
			fprintf(stderr, "FAILED to create config.txt! Program exiting... sorry!\n");
			exit(1);
		}
		
		// TODO: function/macro save config... no copypaste...
		fprintf(config, "theme default\nshow_time 1\nshow_pct 1\nshow_msgs 1\nshow_status 1\nbeep_when_done 1\nproc_clock 100\nauto_brightness 1\nfaded_brightness 5\nbright_pulse 1");
		fflush(config);
		fseek(config, 0, SEEK_SET);
	}
	
	while (!feof(config)) {
		char str[64];
		fscanf(config, "%s", str);
		
		// looks hacky, I know :P
		CHECKSETPROPERTY(config, str, theme, "%s"); // A warning here! Too lazy to fix...
		CHECKSETPROPERTY(config, str, show_time, "%d");
		CHECKSETPROPERTY(config, str, show_pct, "%d");
		CHECKSETPROPERTY(config, str, show_msgs, "%d");
		CHECKSETPROPERTY(config, str, show_status, "%d");
		CHECKSETPROPERTY(config, str, beep_when_done, "%d");
		CHECKSETPROPERTY(config, str, proc_clock, "%d");
		CHECKSETPROPERTY(config, str, auto_brightness, "%d");
		CHECKSETPROPERTY(config, str, faded_brightness, "%d");
		CHECKSETPROPERTY(config, str, bright_pulse, "%d");
		
	}
	
	fclose(config);
	config = NULL;
	
	// battinfo file
	
	battinfo = fopen("battinfo.txt", "r");
	if (battinfo == NULL) {
		fprintf(stderr, "File battinfo.txt not found! Creating one.\n");
		battinfo = fopen("battinfo.txt", "w+");
		
		if (battinfo == NULL) {
			fprintf(stderr, "FAILED to create battinfo.txt! Program exiting... sorry!\n");
			exit(1);
		}
		
		fprintf(battinfo, "battery_min 3000\nbattery_max 4100\nbattery_calibrate 1");
		fflush(battinfo);
		fseek(battinfo, 0, SEEK_SET);
	}
	
	while (!feof(battinfo)) {
		char str[64];
		fscanf(battinfo, "%s", str);
		
		CHECKSETPROPERTY(battinfo, str, battery_min, "%d");
		CHECKSETPROPERTY(battinfo, str, battery_max, "%d");
		CHECKSETPROPERTY(battinfo, str, battery_calibrate, "%d");
	}
	
	fclose(battinfo);
	battinfo = NULL;
	
	
	// theme file
	
	char path[128];
	strcpy(path, theme);
	strcat(path, "/theme.txt");
	
	themefile = fopen(path, "r");
	if (themefile == NULL) {
		fprintf(stderr, "Theme file %s not found! Program exiting... sorry!\n", theme);
		exit(1);
	}
	
	
	while (!feof(themefile)) {
		char str[64];
		fscanf(themefile, "%s", str);

		// looks hacky, I know :P
		CHECKSETPROPERTYPOINT(themefile, str, bar_pos);
		CHECKSETPROPERTYPOINT(themefile, str, completed_pos);
		CHECKSETPROPERTYPOINT(themefile, str, percent_pos);
		CHECKSETPROPERTYPOINT(themefile, str, msg1_pos);
		CHECKSETPROPERTYPOINT(themefile, str, msg2_pos);
		CHECKSETPROPERTYPOINT(themefile, str, status_pos);
		CHECKSETPROPERTYPOINT(themefile, str, time_pos);
		
		if (EQUALS(str, "text_color")) {
			int r, g, b;
			fscanf(themefile, "%d %d %d", &r, &g, &b);
			text_color.r = r;
			text_color.g = g;
			text_color.b = b;
			//continue;
		}
		
		if (EQUALS(str, "text_shadow")) {
			int r, g, b;
			fscanf(themefile, "%d %d %d", &r, &g, &b);
			text_shadow.r = r;
			text_shadow.g = g;
			text_shadow.b = b;
			//continue;
		}
		
		CHECKSETPROPERTY(themefile, str, text_size, "%d");

	}
	
	fclose(themefile);
	themefile = NULL;

	/************************/
	/* initialization stuff */
	
	if ( SDL_Init(initflags) < 0 ) {
		fprintf(stderr, "Couldn't initialize SDL: %s\n",
			SDL_GetError());
		exit(3);
	}

	screen = SDL_SetVideoMode( video_w, video_h, video_bpp, video_flags);

	atexit( SDL_Quit );
	
    if ( screen == NULL ) {
		fprintf(stderr, "Couldn't set %dx%dx%d video mode: %s\n",
                        video_w, video_h, video_bpp, SDL_GetError());
		exit(4);
	}
	
	SDL_FillRect(screen, NULL, 0); // Paint it Black!
	
	SDL_ShowCursor(SDL_DISABLE);
	
	if( TTF_Init() == -1 ) {
		fprintf(stderr, "Couldn't initialize SDL TTF: %s\n",
				TTF_GetError());
		exit(5);
	} 
	
	
	//Initialize SDL_mixer
    if( Mix_OpenAudio( 22050, MIX_DEFAULT_FORMAT, 2, 4096 ) == -1 )
    {
        fprintf(stderr, "Couldn't initialize SDL MIXER: %s\n",
				SDL_GetError());
		
		sound_enabled = FALSE;
		set_vol(0);
		// TODO: Testar se isso evita o ruido.
		//exit(5); 
    }
	
	/*******************/
	/* images & sounds */
	
	strcpy(path, theme);
	strcat(path, "/bg.png");
	bg_surf = loadImage(path, 0, 0);
	
	if (bg_surf == NULL) {
		fprintf(stderr, "Couldn't find the %s image file: %s\n",
				path, SDL_GetError());
		exit(6);
	}
	
	strcpy(path, theme);
	strcat(path, "/fill.png");
	fill_surf = loadImage(path, 0, 0 );
	//fill_surf = loadImage("fill.png", 1, RGB(255,0,255) );
	
	if (fill_surf == NULL) {
		fprintf(stderr, "Couldn't find the %s image file: %s\n",
				path, SDL_GetError());
		exit(6);
	}

	strcpy(path, theme);
	strcat(path, "/completed.png");
	completed_surf = loadImage(path, 0, 0 );
	
	if (completed_surf == NULL) {
		fprintf(stderr, "Couldn't find the %s image file: %s\n",
				path, SDL_GetError());
		exit(6);
	}
	
	if (sound_enabled) {
		strcpy(path, theme);
		strcat(path, "/beep.wav");
		beep = Mix_LoadWAV(path);
		
		// If it cant find the sound file, just ignore.
		if (beep == NULL) {
			fprintf(stderr, "Couldn't find the %s sound file: %s\n",
					path, SDL_GetError());
			//exit(6);
			// Sound not found, continue anyway...
			
			sound_enabled = FALSE;
			Mix_CloseAudio();
		}
	}
	
	
	/**************/
	/* font stuff */
	
	strcpy(path, theme);
	strcat(path, "/font.ttf");
	
	font = TTF_OpenFont( path, text_size ); 

	if( font == NULL ) {
		fprintf(stderr, "Couldn't find the %s font file: %s\n",
				path, SDL_GetError());
		exit(6);
	}
	
	//TTF_SetFontHinting(font, TTF_HINTING_NONE); // Better for retro-styled fonts
	// Not working in dingux... it uses a old SDL_TTF version,
	//  but it looks like TTF_HINTING_NONE by default, so...
	
	msg1 = TTF_RenderText_Solid( font, "Press Select to Exit/ Start hide this", text_color );
	msg1_shadow = TTF_RenderText_Solid( font, "Press Select to Exit/ Start hide this", text_shadow );
	msg2 = TTF_RenderText_Solid( font, "Shoulder + Face Buttons to toggle stuff", text_color );
	msg2_shadow = TTF_RenderText_Solid( font, "Shoulder + Face Buttons to toggle stuff", text_shadow );
	time = TTF_RenderText_Solid( font, "00:00:00", text_color );
	time_shadow = TTF_RenderText_Solid( font, "00:00:00", text_shadow );
	status = TTF_RenderText_Solid( font, "Charging...", text_color );
	status_shadow = TTF_RenderText_Solid( font, "Charging...", text_shadow );
	percent = TTF_RenderText_Solid( font, "  0%", text_color );
	percent_shadow = TTF_RenderText_Solid( font, "  0%", text_shadow );
	

	// Underclock to waste less energy...
	set_clock(proc_clock);
	
	
	// preparations
	int startTicks = SDL_GetTicks();
	int timerTicks = -timer_interval; // start right now...
	int brightTicks = 0; //-brightness_timer_interval;
	
	/*****************/
	/* the main loop */
	
	// Get the button's state easily
	Uint8 *key = SDL_GetKeyState(NULL);
	
	while ( !done ) {
		
		int currentTicks = SDL_GetTicks();
		
		// Every second
		if ( (currentTicks - timerTicks) > timer_interval) {
			int diff = currentTicks - startTicks;
			timerTicks = currentTicks; // update the timer
			
#if !defined(__APPLE__) //testing...
			battery_level = get_battery_level();
#endif
			
			if (battery_calibrate) {
				// Uncomment this to allow it to change the lower value when calibrating...
				// does not "feel" correct...
				//if ((battery_level + TOLERANCE_RANGE) < battery_min ) {
//					battery_min = (battery_level + TOLERANCE_RANGE);
//					printf("min : %d\t", battery_min);
//				}
				if ((battery_level - TOLERANCE_RANGE) > battery_max ) {
					battery_max = (battery_level - TOLERANCE_RANGE);
					printf("max : %d\t", battery_max);
				}
			}
			
			battery_percent = ( (float)(battery_level - battery_min) / (battery_max - battery_min) ) * 100 ;
			battery_percent = CLAMP(battery_percent, 0, 100);
			
			if (battery_percent != battery_percent_old) {
				battery_percent_old = battery_percent;
				battery_percent_changed = TRUE;
			}
			
			char timetext[11];
			sprintf(timetext, "%2.2d:%2.2d:%2.2d", (diff/3600000), (diff/60000)-(60*(diff/3600000)),
					(diff/1000) - (60*(diff/60000)) );
			
			SDL_FreeSurface(time);
			SDL_FreeSurface(time_shadow);
			time = TTF_RenderText_Solid( font, timetext, text_color );
			time_shadow = TTF_RenderText_Solid( font, timetext, text_shadow );
		}
		
		// Every 10 seconds
		if ( (currentTicks - brightTicks) > brightness_timer_interval) {
			brightTicks	= currentTicks; // update the timer
			
			// Fade if necessary
			if ((bright_pulse && !brightness_fading) || auto_brightness) {
				brightness_fading = TRUE;
			}
		}
		
		// The brightness fading animation
		// If fade effect is off, just turn down the brightness
		if (bright_pulse && brightness_fading) {
			static float effect = 0.0;
			effect += 0.1;
			current_brightness = (int)( 11.f + ( 10.f * cos(effect) ) );
			
			// Prevent strange stuff
			if (effect >= 6.28318 || effect < 0.0) { // 2PI
				effect = 0.0;
			}
			
			//printf("bright: %d, ef: %f cos_ef: %f\n", current_brightness, effect, cos( effect ));
		} else {
			if (brightness_fading) {
				current_brightness = faded_brightness;
			} else {
				current_brightness = default_brightness;
			}
		}
		
		//======================================================================

		SDL_UpperBlit(bg_surf, NULL, screen, NULL);
				
		SDL_Rect fillbar = { 0, 0, (fill_surf->w)*(battery_percent/100.f), fill_surf->h };

		applyBitmap(bar_pos.x, bar_pos.y, fill_surf, screen, &fillbar);
		
		if (battery_percent >= 100) {
			// Draw the "completed" pic
			applyBitmap(completed_pos.x, completed_pos.y, completed_surf, screen, NULL);
		}
			
		// Update battery visualization
		if (battery_percent_changed) {
			battery_percent_changed = FALSE;
			
			char batttext[6];
			sprintf(batttext, "%3d%%", battery_percent);
			
			SDL_FreeSurface(percent);
			SDL_FreeSurface(percent_shadow);
			percent = TTF_RenderText_Solid( font, batttext, text_color );
			percent_shadow = TTF_RenderText_Solid( font, batttext, text_shadow );
			
			SDL_FreeSurface(status);
			SDL_FreeSurface(status_shadow);
			
			// Okay! It looks charged!
			if (battery_percent >= 100) {
				status = TTF_RenderText_Solid( font, "Charged!", text_color );
				status_shadow = TTF_RenderText_Solid( font, "Charged!", text_shadow );
				
				// It may beep more than one time...
				if (beep_when_done && sound_enabled) {
					set_vol(default_volume); //TEST
					Mix_PlayChannel( -1, beep, 0 );
				}
				
			} else {
				
				// NOTICE:
				// Please note that it does not actually check if the dingoo is
				//  charging, nor if the usb cable is plugged in.
				//
				// To do it you just need to read some GPIO pins,
				//  it is easy, but I dont feel like doing it now,
				//  feel free to fix this.
				
				status = TTF_RenderText_Solid( font, "Charging...", text_color );
				status_shadow = TTF_RenderText_Solid( font, "Charging...", text_shadow );
			}

		}
		
		if (show_msgs) {
			// center horizontaly if -1
			applyBitmap( (msg1_pos.x == -1) ? CENTER(msg1) : msg1_pos.x+1 , msg1_pos.y+1, msg1_shadow, screen, NULL);
			applyBitmap( (msg1_pos.x == -1) ? CENTER(msg1) : msg1_pos.x , msg1_pos.y, msg1, screen, NULL);
			
			applyBitmap( (msg2_pos.x == -1) ? CENTER(msg2) : msg2_pos.x+1 , msg2_pos.y+1, msg2_shadow, screen, NULL);
			applyBitmap( (msg2_pos.x == -1) ? CENTER(msg2) : msg2_pos.x , msg2_pos.y, msg2, screen, NULL);
			
		}

		if (show_status) {
			applyBitmap( (status_pos.x == -1) ? CENTER(status) : status_pos.x+1 , status_pos.y+1 , status_shadow, screen, NULL);
			applyBitmap( (status_pos.x == -1) ? CENTER(status) : status_pos.x , status_pos.y, status, screen, NULL);
		}

		if (show_pct) {
			applyBitmap( (percent_pos.x == -1) ? CENTER(percent)	: percent_pos.x+1, percent_pos.y+1,
						percent_shadow, screen, NULL);
			applyBitmap( (percent_pos.x == -1) ? CENTER(percent)	: percent_pos.x, percent_pos.y,
						percent, screen, NULL);
		}
		
		if (show_time) {
			applyBitmap( (time_pos.x == -1) ? CENTER(time)	: time_pos.x+1, time_pos.y+1,
						time_shadow, screen, NULL);
			applyBitmap( (time_pos.x == -1) ? CENTER(time)	: time_pos.x, time_pos.y,
						time, screen, NULL);
		}
		
		//======================================================================
		
		while ( SDL_PollEvent(&e) )
		{
			switch (e.type)
			{
				// pressed some key...
				case SDL_KEYDOWN:
					
					// dont fire any action if it was fading before
					if (!brightness_fading) {
						switch (e.key.keysym.sym)
						{
							case SDLK_ESCAPE:
								done = 1;
								break;
							case SDLK_RETURN:
								show_msgs = !show_msgs; // toggle help messages
								break;
								
							// Press one of the shoulder buttons + the A, B, X or Y buttons
							case DINGOO_A: // button A pressed
								if (key[DINGOO_R] || key[DINGOO_L]) // any shoulder button pressed
									show_status = !show_status;
								break;
							case DINGOO_B:
								if (key[DINGOO_R] || key[DINGOO_L])
									show_time = !show_time;
								break;
							case DINGOO_Y:
								if (key[DINGOO_R] || key[DINGOO_L])
									show_pct = !show_pct;
								break;
							case DINGOO_X:
								if (key[DINGOO_R] || key[DINGOO_L])
									bright_pulse = !bright_pulse;
								break;
								
							default:
								break;
						}
					}
					
					// reset brightness
					brightTicks	= currentTicks;
					brightness_fading = FALSE;
					
					break;
					
				case SDL_QUIT:
					done = 1;
					break;
				default:
					break;
			}
			
		}
		
		// TESTING UI in my computer
#if defined(__APPLE__)
		if (key[SDLK_UP]) {
			battery_level += 7;
			battery_percent_changed = TRUE;
		}
		if (key[SDLK_DOWN]) {
			battery_level -= 7;
			battery_percent_changed = TRUE;
		}
#endif
		
		SDL_Flip( screen );
		
		// set the brightness, so it can animate if necessary
		set_brightness(current_brightness);
		
		// Let it rest a little bit...
		SDL_Delay(30);
	}
	
	// set the clock back to the old value
	set_clock(old_clock);
	
	// SAVE modifications
	config = fopen("config.txt", "w");
	if (config != NULL) {
		fprintf(config, "theme %s\nshow_time %d\nshow_pct %d\nshow_msgs %d\nshow_status %d\nbeep_when_done %d\nproc_clock %d\nauto_brightness %d\nfaded_brightness %d\nbright_pulse %d", theme, show_time, show_pct, show_msgs, show_status, beep_when_done, proc_clock, auto_brightness, faded_brightness, bright_pulse);
		
		fclose(config);
	}
	
	battinfo = fopen("battinfo.txt", "w");
	if (battinfo != NULL) {
		if (battery_calibrate && battery_percent >= 99) {
			battery_calibrate = FALSE;
		}
		fprintf(battinfo, "battery_min %d\nbattery_max %d\nbattery_calibrate %d", battery_min, battery_max, battery_calibrate);
		
		fclose(battinfo);
	}
	
	
	/***************/
	/* clean stuff */
	
	SDL_FreeSurface( msg1 );
	SDL_FreeSurface( msg1_shadow );
	SDL_FreeSurface( msg2 );
	SDL_FreeSurface( msg2_shadow );
	SDL_FreeSurface( time );
	SDL_FreeSurface( time_shadow );
	SDL_FreeSurface( status );
	SDL_FreeSurface( status_shadow );
	SDL_FreeSurface( percent );
	SDL_FreeSurface( percent_shadow );
	
	if (sound_enabled) {
		Mix_FreeChunk( beep );
		Mix_CloseAudio();
	}
	
	TTF_CloseFont( font );
	TTF_Quit(); 
	
	return(0);
}
